#include "scripts/hole-config.lua"
#include "scripts/default-values.lua"

local holeTemplateConfig = holeTemplateConfig
_G.holeTemplateConfig = nil

holeHandler = {
	holeNum = 1,
	holes = {}
}



function checkSetting(name, value, type)
	if not HasKey(name) then
		print(name)
		if type == "f" then
			SetFloat(name, value)
		elseif type == "b" then
			SetBool(name, value)
		elseif type == "s" then
			SetString(name, value)
		end
	end
end



function loadConfig()
	local holes = {}
	for i=1, #holeTemplateConfig do
		local firsthole = holeTemplateConfig[i]
		firsthole.sound = LoadLoop("MOD/snd/"..firsthole.name..".ogg", 15)
		holes[firsthole.name] = firsthole
	end
	return holes
end

function setParticleStyle()
	ParticleReset()
	ParticleType("plain")
	ParticleTile(4)
	ParticleRadius(0.03)
	ParticleAlpha(1)
	ParticleDrag(0)
	ParticleEmissive(1)
	ParticleStretch(10)
	ParticleSticky(0)
	ParticleCollide(0)
end

function setShootStyle(currentHole)
	ParticleReset()
	ParticleType("plain")
	ParticleTile(4)
	ParticleRadius(deleteRadius*0.5, deleteRadius*0.02)
	ParticleDrag(0)
	ParticleEmissive(1, 0.5)
	ParticleSticky(0)
	ParticleCollide(0)
end



function init()
	RegisterTool("omegacube", "Omega Cube", "MOD/vox/omegacube.vox")
	SetBool("game.tool.omegacube.enabled", true)
	SetFloat("game.tool.omegacube.ammo", 101)

	checkSetting("savegame.mod.showinfo", defaultValues["showInfo"], "b")
	checkSetting("savegame.mod.effectplayer", defaultValues["effectPlayer"], "b")
	checkSetting("savegame.mod.destroyclose", defaultValues["destroyClose"], "b")
	checkSetting("savegame.mod.growball", defaultValues["growBall"], "b")
	checkSetting("savegame.mod.maxtime", defaultValues["maxTime"], "f")
	checkSetting("savegame.mod.maxdist", defaultValues["maxDistance"], "f")
	checkSetting("savegame.mod.startmass", defaultValues["startMass"], "f")
	checkSetting("savegame.mod.maxmass", defaultValues["maxMass"], "f")
	checkSetting("savegame.mod.growspeed", defaultValues["growSpeed"], "f")
	checkSetting("savegame.mod.deletemass", defaultValues["deleteMass"], "f")
	checkSetting("savegame.mod.deletedist", defaultValues["deleteRadius"], "f")
	checkSetting("savegame.mod.menukey", defaultValues["menuKey"], "s")
	checkSetting("savegame.mod.killkey", defaultValues["killKey"], "s")
	checkSetting("savegame.mod.holename", defaultValues["holeName"], "s")
	checkSetting("savegame.mod.fieldParticles", defaultValues["fieldParticles"], "b")
	checkSetting("savegame.mod.shotTrailLen", defaultValues["shotTrailLen"], "f")
	checkSetting("savegame.mod.shotSpeed", defaultValues["shotTrailLen"], "f")
	checkSetting("savegame.mod.frequency", defaultValues["frequency"], "f")

	holeTemplates = loadConfig()
	
	showInfo = GetBool("savegame.mod.showinfo")
	effectPlayer = GetBool("savegame.mod.effectplayer")
	destroyClose = GetBool("savegame.mod.destroyclose")
	growBall = GetBool("savegame.mod.growball")
	maxTime = GetFloat("savegame.mod.maxtime")
	maxDistance = GetFloat("savegame.mod.maxdist")
	startMass = GetFloat("savegame.mod.startmass")
	maxMass = GetFloat("savegame.mod.maxmass")
	growSpeed = GetFloat("savegame.mod.growspeed")
	deleteMass = GetFloat("savegame.mod.deletemass")
	deleteRadius = GetFloat("savegame.mod.deletedist")
	menuKey = GetString("savegame.mod.menukey")
	killKey = GetString("savegame.mod.killkey")
	holeName = GetString("savegame.mod.holename")
	fieldParticles = GetBool("savegame.mod.fieldParticles")
	shotTrailLen = GetFloat("savegame.mod.shotTrailLen")
	shotSpeed = GetFloat("savegame.mod.shotSpeed")
	frequency = GetFloat("savegame.mod.frequency")

	optionsOpen = false

	angle = 10

	throwSound = LoadSound("MOD/snd/bram.ogg", 2)
	impactSound = LoadSound("MOD/snd/hit.ogg", 15)
	toggleSound = LoadSound("MOD/snd/toggle.ogg")
	menuOpenSound = LoadSound("MOD/snd/menu-open.ogg")
	menuCloseSound = LoadSound("MOD/snd/menu-close.ogg")
	gamma = 6.6743*10^-4
	particlesPerS = 500
	particleCameraRadius = 10
	basisLight = 3
	shootLight = 80
	cubeLight = basisLight
	glareSprite = LoadSprite("gfx/glare.png")

	onColor = {0, 1, 0}
	offColor = {0.75, 0.75, 0.75}
end


function handleCommand(cmd)
	if cmd == "quickload" then
		holeTemplates = loadConfig()
		for key, currentHole in ipairs(holeHandler.holes) do
			currentHole.gravityDirection = holeTemplates[currentHole.name].gravityDirection
		end
	end
end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end

function GetAimPos()
	local ct = GetCameraTransform()
	local forwardPos = TransformToParentPoint(ct, Vec(0, 0, -100))
    local direction = VecSub(forwardPos, ct.pos)
    local distance = VecLength(direction)
	local direction = VecNormalize(direction)
	local hit, hitDistance = QueryRaycast(ct.pos, direction, distance)
	if hit then
		forwardPos = TransformToParentPoint(ct, Vec(0, 0, -hitDistance))
		distance = hitDistance
	end
	return forwardPos, hit, distance
end

function hold(dt, toolName)
	if cubeLight > basisLight then
		cubeLight = math.max(basisLight, cubeLight*(1-dt))
	end
	holdingTheCube = GetString("game.player.tool") == toolName and GetPlayerVehicle() == 0
	if InputPressed(menuKey) then
		SetString("game.player.tool", toolName)
	end

	if holdingTheCube then

		if InputPressed("lmb") and not optionsOpen then
			Shoot(holeName)
		end

		if InputPressed(menuKey) or InputPressed("rmb") then
			if optionsOpen then
				PlaySound(menuCloseSound)
			else
				PlaySound(menuOpenSound)
			end
			optionsOpen = not optionsOpen 
		end

		local b = GetToolBody()
		if b ~= 0 then
			angle = angle + dt * 300
			local bt = GetBodyTransform(b)
			worldToolCenter = TransformToParentPoint(bt, Vec(-0.025, 0.025, -0.025))
			local toolCenter = TransformToLocalVec(GetPlayerTransform(), VecSub(worldToolCenter, bt.pos))		
			SetToolTransform(Transform(VecSub(Vec(0.65, -0.35, -1.0), toolCenter), QuatEuler(angle/3, angle/5, angle/2)))
			PointLight(worldToolCenter, holeTemplates[holeName].r, holeTemplates[holeName].g, holeTemplates[holeName].b, cubeLight)
		end
	end
end

function Shoot(holeName)
	local aimpos, hit, distance = GetAimPos()
	local gunpos = VecCopy(worldToolCenter)

	holeHandler.holes[holeHandler.holeNum] = deepcopy(holeTemplates[holeName])
	local currentHole = holeHandler.holes[holeHandler.holeNum]
	currentHole.inair = true
	currentHole.dist = 0
	currentHole.timeToLive = maxTime
	currentHole.pos = gunpos
	currentHole.shotSpeed = shotSpeed
	currentHole.shotTrailLen = shotTrailLen
	currentHole.shotDirection = VecNormalize(VecSub(aimpos, gunpos))

	holeHandler.holeNum = holeHandler.holeNum + 1
	PointLight(worldToolCenter, currentHole.r, currentHole.g, currentHole.b, shootLight)
	cubeLight = shootLight
	PlaySound(throwSound, worldToolCenter, 0.5)
end

function HoleOperations(dt, currentHole)
	local point2 = VecAdd(currentHole.pos, VecScale(currentHole.shotDirection, currentHole.shotSpeed * dt))
	QueryRequire("physical")
	local hit, dist = QueryRaycast(currentHole.pos, currentHole.shotDirection, VecLength(VecSub(point2, currentHole.pos)))
	if hit then
		currentHole.showInfo = showInfo
		currentHole.effectPlayer = effectPlayer
		currentHole.destroyClose = destroyClose
		currentHole.growBall = growBall
		currentHole.maxTime = maxTime
		if maxDistance > 499 then
			currentHole.maxDistance = 1000000
		else
			currentHole.maxDistance = maxDistance
		end
		currentHole.maxMass = maxMass
		currentHole.growSpeed = growSpeed
		currentHole.deleteMass = deleteMass
		currentHole.deleteRadius = deleteRadius
		
		currentHole.pos = VecAdd(currentHole.pos, VecScale(VecNormalize(VecSub(point2, currentHole.pos)), dist))
		currentHole.inair = false
		currentHole.mass = startMass
		currentHole.dist = 0
		currentHole.initialDeleteRadius = deleteRadius
		currentHole.timeToLive = maxTime
		currentHole.frequency = frequency

		MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
		spawnBall(currentHole)
		PlaySound(impactSound, currentHole.pos, deleteRadius)
	else
		PointLight(point2, currentHole.r, currentHole.g, currentHole.b, 30)
		local spriteTF = Transform(point2, GetCameraTransform().rot)
		DrawSprite(glareSprite, spriteTF, 20, 20, currentHole.r, currentHole.g, currentHole.b, 1, true, true)
		DrawSprite(glareSprite, spriteTF, 10, 10, 1, 1, 1, 1, true, true)
		setShootStyle()
		ParticleColor(1, 1, 1, currentHole.r, currentHole.g, currentHole.b)
		for i=1, math.max(round(currentHole.shotSpeed, 0), 1) do
			SpawnParticle(VecLerp(currentHole.pos, point2, i/currentHole.shotSpeed), Vec(0,0,0), math.min(currentHole.shotTrailLen/currentHole.shotSpeed, 120))
		end
		currentHole.pos = point2
		PlayLoop(currentHole.sound, point2, 0.1)
	end
end

function spawnBall(currentHole)
	currentHole.ball = Spawn(string.format('<vox density="0" tag="unbreakable dynamic=false physical=false" physical="false" dynamic="false" collide="false" file="MOD/vox/holes/%s.vox" scale="%.2f"/>', currentHole.name, currentHole.deleteRadius/5), Transform(currentHole.pos), true)[1]
end

function replaceBall(currentHole)
	local b = currentHole.ball
	if IsHandleValid(b) then
		Delete(b)
	end
	spawnBall(currentHole)
end

function Hole(dt, currentHole)
	if currentHole.effectPlayer then
		GravityEffectPlayer(dt, currentHole)
	end
	if currentHole.destroyClose then
		RipAndTear(currentHole)
	end
	tickBall(currentHole)

	local mi = VecAdd(currentHole.pos, Vec(-currentHole.dist, -currentHole.dist, -currentHole.dist))
	local ma = VecAdd(currentHole.pos, Vec(currentHole.dist, currentHole.dist, currentHole.dist))
	QueryRequire("physical dynamic")
	local bodies = QueryAabbBodies(mi, ma)
	for i=1,#bodies do
		local b = bodies[i]
		local mass = GetBodyMass(b)
		if mass < currentHole.deleteMass then
			Delete(b)
			currentHole.mass = math.min(maxMass, currentHole.mass + mass)
		else
			local bc = TransformToParentPoint(GetBodyTransform(b), GetBodyCenterOfMass(b))
			local dir = VecSub(currentHole.pos, bc)
			local dist = VecLength(dir)
			if dist < currentHole.deleteRadius then
				Delete(b)
				currentHole.mass = math.min(currentHole.maxMass, currentHole.mass + mass)
			elseif dist < currentHole.dist then
				local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole, b)
				local impulse = VecScale(gdir, dt * amplitude * gamma * mass * currentHole.mass/(dist * dist * dist))
				ApplyBodyImpulse(b, bc, impulse)
			end
		end
	end
end



function updateBallSize(currentHole, sizeFactor)
	currentHole.deleteRadius = math.min(currentHole.initialDeleteRadius * sizeFactor, 100)
	MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
	replaceBall(currentHole)
end


function tickBall(currentHole)
	if currentHole.growBall then
		if currentHole.sizeFactor * currentHole.initialDeleteRadius > currentHole.deleteRadius * 1.15 or currentHole.sizeFactor * currentHole.initialDeleteRadius < currentHole.deleteRadius then
			updateBallSize(currentHole, currentHole.sizeFactor)
		end
	else
		if currentHole.initialDeleteRadius ~= currentHole.deleteRadius then
			updateBallSize(currentHole, 1)
		end
	end
end


function GetRandomVector(OffsetRadius, currentHoleWidth)
	local lenVec = Vec(0, math.random()*currentHoleWidth+OffsetRadius, 0)
	return QuatRotateVec(QuatEuler(math.random(360), math.random(360), math.random(360)), lenVec)
end

function RipAndTear(currentHole)
	ripAmount = currentHole.sizeFactor * 0.1
	if math.random() < ripAmount then
		for x=0, ripAmount do
			Rip(currentHole)
		end
	end
end

function Rip(currentHole)
	local rippingRadius = math.min(maxDistance-1, math.sqrt(currentHole.mass * gamma/100))
	if rippingRadius > currentHole.deleteRadius then
		local holePosition = VecAdd(currentHole.pos, GetRandomVector(currentHole.deleteRadius, rippingRadius-currentHole.deleteRadius))
		local scaledGForce = math.min(math.abs(GetGForce(currentHole, holePosition)), 400)
		MakeHole(holePosition, scaledGForce, math.max(0, 0.75*scaledGForce-15), math.max(0, 0.5*scaledGForce-20))
	end
end

function GravityEffectPlayer(dt, currentHole)
	local pt = GetPlayerPos()
	local dir = VecSub(currentHole.pos, pt)
	local dist = VecLength(dir)
	if dist < currentHole.dist then
		if dist < 0.4*currentHole.deleteRadius then
			RespawnPlayer()
		else
			if dist < currentHole.deleteRadius then
				SetPlayerHealth(math.min(GetPlayerHealth()-0.1*dist/currentHole.deleteRadius, 0))
			end
			local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
			local add = VecScale(gdir, dt * amplitude * gamma * currentHole.mass/(dist * dist * dist))
			local pv = GetPlayerVelocity()
			SetPlayerVelocity(VecAdd(pv, add))
		end
	end
end

function GetTotalPlayerG()
	local playerPos = GetPlayerPos()
	local playerG = Vec(0, -10, 0)
	for key, currentHole in ipairs(holeHandler.holes) do
		if not currentHole.inair and currentHole.timeToLive > 0 then
			local dir = VecSub(currentHole.pos, playerPos)
			local dist = VecLength(dir)
			
			if dist < currentHole.dist then
				local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
				playerG = VecAdd(playerG, VecScale(gdir, amplitude * gamma * currentHole.mass/(dist * dist * dist)))
			end
		end
	end
	return playerG
end



function endHole(currentHole)
	MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
	if IsHandleValid(currentHole.ball) then
		Delete(currentHole.ball)
	end
	if currentHole.timeToLive < 0.01 then
		if currentHole.showInfo then
			SetString("hud.notification", (currentHole.name:gsub("^%l", string.upper)).." Vaporized")
		end
		Explosion(currentHole.pos, 4)
	else
		if currentHole.showInfo then
			SetString("hud.notification", (currentHole.name:gsub("^%l", string.upper)).." Removed")
		end
	end
	currentHole.timeToLive = 0
end

function GetGForce(currentHole, position)
	local dir = VecSub(currentHole.pos, position)
	local dist = VecLength(dir)
	
	if dist < currentHole.dist then
		local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
		return VecLength(gdir) * amplitude * gamma * currentHole.mass/(dist * dist * dist * 10)
	else
		return 0
	end
end

function comma_value(amount)
	local formatted, k = amount, 0
	while true do  
		formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
		if (not k) or k==0 then
		  break
		end
	end
	return formatted
end


function getParticleCoords(dt)
	local cameraTf = GetCameraTransform()
	local newParticleNum = math.floor(particlesPerS*dt)
	local particleCoords = {}
	for i=1, newParticleNum do
		particleCoords[i] = {}
		particleCoords[i].pos = TransformToParentPoint(cameraTf, Vec(math.random(-particleCameraRadius*2, particleCameraRadius*2), math.random(-particleCameraRadius/2, particleCameraRadius), math.random(-particleCameraRadius*3, particleCameraRadius/2)))
		particleCoords[i].vel = Vec(0, 0, 0)
	end
	return particleCoords
end

function applyParticleVelocity(dt, particleCoords)	
	for key, currentHole in ipairs(holeHandler.holes) do
		if currentHole.timeToLive > 0.01 then
			for key, particle in ipairs(particleCoords) do
				local dir = VecSub(currentHole.pos, particle.pos)
				local dist = VecLength(dir)
				if dist < currentHole.dist then
					if dist < currentHole.deleteRadius then
						particle = nil
					else
						local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
						particle.vel = VecAdd(particle.vel, VecScale(gdir, 20 * dt * amplitude * gamma * currentHole.mass/(dist * dist * dist)))
					end
				end
			end
		end
	end
	return particleCoords
end

function spawnParticles(dt, particleCoords)
	setParticleStyle()
	for i=1, #particleCoords do
		local totalVel = VecLength(particleCoords[i].vel)
		if 0.01 < totalVel then
			totalVel = totalVel/10
			ParticleEmissive(totalVel)
			ParticleColor(totalVel, totalVel, totalVel)
			SpawnParticle(particleCoords[i].pos, particleCoords[i].vel, math.max(math.min(1/totalVel, 0.3)), dt)
		end
	end
end

function tickParticles(dt)
	local particleCoords = getParticleCoords(dt)
	particleCoords = applyParticleVelocity(dt, particleCoords)
	spawnParticles(dt, particleCoords)
end


function tick(dt)
	hold(dt, "omegacube")
	for key, currentHole in ipairs(holeHandler.holes) do
		if currentHole.inair then
			HoleOperations(dt, currentHole)
		else
			if currentHole.timeToLive > 0 then
				currentHole.timeToLive = math.min(currentHole.timeToLive - dt, currentHole.maxTime)
				if InputPressed(killKey) or currentHole.timeToLive < 0.01 then
					endHole(currentHole)
					holeHandler.holes[key] = {timeToLive=0}
				else
					currentHole.mass = math.min(currentHole.maxMass, currentHole.mass * (1+dt*0.00727*currentHole.growSpeed))
					currentHole.dist = math.min(currentHole.maxDistance, math.sqrt(gamma * currentHole.mass/0.5))
					currentHole.sizeFactor = (20*math.pow(currentHole.mass, 1/2)/math.pow(2100000000, 1/2))+1
					currentHole.sinFactor = math.sin(currentHole.timeToLive*currentHole.frequency*math.pi)
					PlayLoop(currentHole.sound, currentHole.pos, 0.5*math.log(currentHole.mass)/math.log(maxMass))
					if currentHole.showInfo then
						SetString("hud.notification", string.format("Mass "..comma_value(round(currentHole.mass/1000, 0)).." t - Gravity %.2f g", GetGForce(currentHole, GetPlayerPos())))
					end
					Hole(dt, currentHole)
				end
			end
		end
	end
	if fieldParticles then
		if #holeHandler.holes > 0 then
			tickParticles(dt)
		end
	end
end


function UiHoleSelector()
	for i=1, #holeTemplateConfig do
		UiHole = holeTemplateConfig[i]
		UiPush()
			if holeName == UiHole.name then
				UiColor(1, 1, 1, 0.3)
				UiImageBox("ui/common/box-solid-6.png", 200, 40, 6, 6)
			end
			UiButtonImageBox("ui/common/box-outline-fill-6.png", 12, 12, UiHole.r, UiHole.g, UiHole.b)
			UiColor(.9,.9,.9)
			if UiTextButton(UiHole.name:gsub("^%l", string.upper), 200, 40) then
				SetString("savegame.mod.holename", UiHole.name)
				PlaySound(toggleSound)
				holeName = UiHole.name
				if #holeHandler.holes > 0 then
					local currentHole = holeHandler.holes[holeHandler.holeNum-1]
					if currentHole.timeToLive > 0 then
						currentHole.name = holeName
						currentHole.gravityDirection = UiHole.gravityDirection
						currentHole.gravityAmplitude = UiHole.gravityAmplitude
						currentHole.r = UiHole.r
						currentHole.g = UiHole.g
						currentHole.b = UiHole.b
						currentHole.sound = UiHole.sound
						if not currentHole.inair then
							replaceBall(currentHole)
						end
					end
				end
			end
		UiPop()
		UiTranslate(0, 50)
	end
	UiTranslate(0, -50)
end

function draw()
	local width = 975
	local height = 875
	local gap = 22
	local currentHole = holeHandler.holes[holeHandler.holeNum-1] or {}
	if holdingTheCube then
		if optionsOpen then
			UiMakeInteractive()
			UiAlign("top left")
			UiTranslate(gap, UiHeight()-height-gap)
			
			UiColor(0,0,0,0.75)
			UiImageBox("ui/common/box-solid-10.png", width, height, 6, 6)
			UiTranslate(gap, gap)
			UiColor(1, 1, 1)
			UiFont("bold.ttf", 48)
			UiText("Omega Holes")
			UiColor(1, 1, 1)
			UiFont("regular.ttf", 26)

			UiTranslate(width-gap*2, 0)
			UiPush()
				UiTranslate(-40, 0)
				UiColor(1, 0, 0)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if UiTextButton("X", 40, 40) then
					optionsOpen = false
					PlaySound(toggleSound)
					PlaySound(menuCloseSound)
				end
			UiPop()
			
			UiPush()
				UiAlign("top right")
				UiTranslate(-50, 0)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if UiTextButton("Reset to Default", 200, 40) then
					SetBool("savegame.mod.showinfo", defaultValues["showInfo"])
					SetBool("savegame.mod.effectplayer", defaultValues["effectPlayer"])
					SetBool("savegame.mod.destroyclose", defaultValues["destroyClose"])
					SetBool("savegame.mod.growball", defaultValues["growBall"])
					SetFloat("savegame.mod.maxtime", defaultValues["maxTime"])
					SetFloat("savegame.mod.maxdist", defaultValues["maxDistance"])
					SetFloat("savegame.mod.startmass", defaultValues["startMass"])
					SetFloat("savegame.mod.maxmass", defaultValues["maxMass"])
					SetFloat("savegame.mod.growspeed", defaultValues["growSpeed"])
					SetFloat("savegame.mod.deletemass", defaultValues["deleteMass"])
					SetFloat("savegame.mod.deletedist", defaultValues["deleteRadius"])
					SetBool("savegame.mod.fieldParticles", defaultValues["fieldParticles"])
					SetFloat("savegame.mod.shotTrailLen", defaultValues["shotTrailLen"])
					SetFloat("savegame.mod.shotSpeed", defaultValues["shotTrailLen"])
					SetFloat("savegame.mod.frequency", defaultValues["frequency"])
					
					showInfo = defaultValues["showInfo"]
					effectPlayer = defaultValues["effectPlayer"]
					destroyClose = defaultValues["destroyClose"]
					growBall = defaultValues["growBall"]
					maxTime = defaultValues["maxTime"]
					maxDistance = defaultValues["maxDistance"]
					startMass = defaultValues["startMass"]
					maxMass = defaultValues["maxMass"]
					growSpeed = defaultValues["growSpeed"]
					deleteMass = defaultValues["deleteMass"]
					deleteRadius = defaultValues["deleteRadius"]
					fieldParticles = defaultValues["fieldParticles"]
					shotTrailLen = defaultValues["shotTrailLen"]
					shotSpeed = defaultValues["shotSpeed"]
					frequency = defaultValues["frequency"]
					PlaySound(toggleSound)
					initialDeleteRadius = deleteRadius
				end
			UiPop()

			
			UiTranslate(-width*0.97+gap*2, 100)
			UiPush()
				UiAlign("left")
				UiText("Show Info")
				UiTranslate(310, 0)
				UiAlign("right")
				UiColor(0.4, 0.9, 1)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if showInfo then
					UiColor(onColor[1], onColor[2], onColor[3])
					if UiTextButton("Yes", 60, 30) then
						showInfo = false
						currentHole.showInfo = showInfo
						SetBool("savegame.mod.showinfo", showInfo)
						PlaySound(toggleSound)
					end
				else
					UiColor(offColor[1], offColor[2], offColor[3])
					if UiTextButton("No", 60, 30) then
						showInfo = true
						currentHole.showInfo = showInfo
						SetBool("savegame.mod.showinfo", showInfo)
						PlaySound(toggleSound)
					end
				end
			UiPop()


			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Show Field Particles")
				UiTranslate(310, 0)
				UiAlign("right")
				UiColor(0.4, 0.9, 1)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if fieldParticles then
					UiColor(onColor[1], onColor[2], onColor[3])
					if UiTextButton("Yes", 60, 30) then
						fieldParticles = false
						currentHole.effectPlayer = effectPlayer
						SetBool("savegame.mod.fieldParticles", fieldParticles)
						PlaySound(toggleSound)
					end
				else
					UiColor(offColor[1], offColor[2], offColor[3])
					if UiTextButton("No", 60, 30) then
						fieldParticles = true
						currentHole.effectPlayer = effectPlayer
						SetBool("savegame.mod.fieldParticles", fieldParticles)
						PlaySound(toggleSound)
					end
				end
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Gravity moves Player")
				UiTranslate(310, 0)
				UiAlign("right")
				UiColor(0.4, 0.9, 1)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if effectPlayer then
					UiColor(onColor[1], onColor[2], onColor[3])
					if UiTextButton("Yes", 60, 30) then
						effectPlayer = false
						currentHole.effectPlayer = effectPlayer
						SetBool("savegame.mod.effectplayer", effectPlayer)
						PlaySound(toggleSound)
					end
				else
					UiColor(offColor[1], offColor[2], offColor[3])
					if UiTextButton("No", 60, 30) then
						effectPlayer = true
						currentHole.effectPlayer = effectPlayer
						SetBool("savegame.mod.effectplayer", effectPlayer)
						PlaySound(toggleSound)
					end
				end
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Crush Close Objects")
				UiTranslate(310, 0)
				UiAlign("right")
				UiColor(0.4, 0.9, 1)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if destroyClose then
					UiColor(onColor[1], onColor[2], onColor[3])
					if UiTextButton("Yes", 60, 30) then
						destroyClose = false
						currentHole.destroyClose = destroyClose
						SetBool("savegame.mod.destroyclose", destroyClose)
						PlaySound(toggleSound)
					end
				else
					UiColor(offColor[1], offColor[2], offColor[3])
					if UiTextButton("No", 60, 30) then
						destroyClose = true
						currentHole.destroyClose = destroyClose
						SetBool("savegame.mod.destroyclose", destroyClose)
						PlaySound(toggleSound)
					end
				end
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Expand the Sphere")
				UiTranslate(310, 0)
				UiAlign("right")
				UiColor(0.4, 0.9, 1)
				UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
				if growBall then
					UiColor(onColor[1], onColor[2], onColor[3])
					if UiTextButton("Yes", 60, 30) then
						growBall = false
						currentHole.growBall = growBall
						SetBool("savegame.mod.growball", growBall)
						PlaySound(toggleSound)
					end
				else
					UiColor(offColor[1], offColor[2], offColor[3])
					if UiTextButton("No", 60, 30) then
						growBall = true
						currentHole.growBall = growBall
						SetBool("savegame.mod.growball", growBall)
						PlaySound(toggleSound)
					end
				end
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Projectile Shot Velocity")
				UiTranslate(450, 10)
				UiAlign("right")
				shotSpeed, done = exponentialSlider(shotSpeed, 0.1, 1000, 0, 0.7, 0.9)
				shotSpeed = round(shotSpeed, 2)
				currentHole.shotSpeed = shotSpeed
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f m/s", shotSpeed))
				SetFloat("savegame.mod.shotSpeed", shotSpeed)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Projectile Trail Length")
				UiTranslate(450, 10)
				UiAlign("right")
				shotTrailLen, done = exponentialSlider(shotTrailLen, 0.1, 1000, 0, 0.7, 0.9)
				shotTrailLen = round(shotTrailLen, 2)
				currentHole.shotTrailLen = shotTrailLen
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f m", shotTrailLen))
				SetFloat("savegame.mod.shotTrailLen", shotTrailLen)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Hole Life Time")
				UiTranslate(450, 10)
				UiAlign("right")
				maxTime, done = exponentialSlider(maxTime, 0.1, 7200, 0.9, 0.7, 0)
				maxTime = round(maxTime, 2)
				currentHole.maxTime = maxTime
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f s", maxTime))
				SetFloat("savegame.mod.maxtime", maxTime)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Maximum Range")
				UiTranslate(450, 10)
				UiAlign("right")
				maxDistance, done = exponentialSlider(maxDistance, 5, 500, 0.9, 0.7, 0)
				maxDistance = round(maxDistance, 0)
				currentHole.maxDistance = maxDistance
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				if maxDistance > 499 then
					UiText("infinity m")
				else
					UiText(maxDistance.." m")
				end
				SetFloat("savegame.mod.maxdist", maxDistance)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Starting Mass")
				UiTranslate(450, 10)
				UiAlign("right")
				startMass, done = exponentialSlider(startMass, 1000, 21000000000, 0.9, 0.7, 0)
				startMass = round(startMass, -2)
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(comma_value(startMass/1000).." t")
				SetFloat("savegame.mod.startmass", startMass)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Maximum Mass")
				UiTranslate(450, 10)
				UiAlign("right")
				maxMass, done = exponentialSlider(maxMass, 1000, 21000000000, 0.9, 0.7, 0)
				maxMass = round(maxMass, -2)
				currentHole.maxMass = maxMass
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(comma_value(maxMass/1000).." t")
				SetFloat("savegame.mod.maxmass", maxMass)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Mass Gain in % per s")
				UiTranslate(450, 10)
				UiAlign("right")
				growSpeed, done = exponentialSlider(growSpeed, 0.25, 1000, 0.9, 0.7, 0)
				growSpeed = round(growSpeed, 2)
				currentHole.growSpeed = growSpeed
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f", growSpeed).." %/s")
				SetFloat("savegame.mod.growspeed", growSpeed)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Hole Radius")
				UiTranslate(450, 10)
				UiAlign("right")
				deleteRadius, done = exponentialSlider(deleteRadius, 0.1, 100, 0.9, 0.7, 0)
				deleteRadius = round(deleteRadius, 2)
				currentHole.initialDeleteRadius = deleteRadius
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f m", deleteRadius))
				SetFloat("savegame.mod.deletedist", deleteRadius)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Swing Frequency")
				UiTranslate(450, 10)
				UiAlign("right")
				frequency, done = exponentialSlider(frequency, 0.01, 10, 0.9, 0.4, 0)
				frequency = round(frequency, 2)
				currentHole.frequency = frequency
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f Hz", frequency))
				SetFloat("savegame.mod.frequency", frequency)
			UiPop()

			UiTranslate(0, 50)
			UiPush()
				UiAlign("left")
				UiText("Removes Debris if Mass <")
				UiTranslate(450, 10)
				UiAlign("right")
				deleteMass, done = exponentialSlider(deleteMass, 0.01, 1000, 0.7, 0.9, 0)
				deleteMass = round(deleteMass, 2)
				currentHole.deleteMass = deleteMass
				UiTranslate(210, 0)
				UiColor(0.9, 0.9, 0.9)
				UiText(string.format("%.2f kg", deleteMass))
				SetFloat("savegame.mod.deletemass", deleteMass)
			UiPop()

			UiTranslate(700, -height+gap*2+110)
			UiHoleSelector()
		else
			miniHeight = 170
			miniWidth = 290
			spacer = 110
			UiAlign("top left")
			UiColor(0,0,0,0.75)
			UiTranslate(gap, UiHeight()-miniHeight-gap)
			UiImageBox("ui/common/box-solid-10.png", miniWidth, miniHeight, 6, 6)
			
			UiTranslate(gap/2, gap/2)
			UiFont("bold.ttf", 30)
			UiColor(1, 1, 1)
			UiText("Controlls")

			UiFont("regular.ttf", 26)
			UiColor(0.9, 0.9, 0.9)

			UiTranslate(0, 34)
			UiPush()
				UiColor(0.9, 0.9, 0.9)
				UiImageBox("ui/common/box-solid-10.png", miniWidth-gap, 32, 6, 6)
				UiTranslate(0, 5)
				UiColor(0, 0, 0)
				UiText("Gravity")

				UiTranslate(miniWidth-gap-5, 0)
				UiAlign("top right")
				if effectPlayer then
					local playerG = GetTotalPlayerG()
					UiText(string.format("%.2f g", round(VecLength(playerG)/10, 2)))
				else
					UiText("1.00 g")
				end
			UiPop()

			UiTranslate(0, 36)
			UiPush()
				UiText("Shoot")
				UiTranslate(spacer, 0)
				UiText("Left-click")
			UiPop()

			UiTranslate(0, 26)
			UiPush()
				UiText("Options")
				UiTranslate(spacer, 0)
				UiText("Right-click or "..menuKey)
			UiPop()

			UiTranslate(0, 26)
			UiPush()
				UiText("Remove")
				UiTranslate(spacer, 0)
				UiText(killKey)
			UiPop()
		end
	end
end

function exponentialSlider(val, min, max, r, g, b)
	UiColor(r, g, b)
	UiPush()
		UiTranslate(0, -8)
		local w = 195
		local done = false
		local base = math.pow(max/min, 1/195)
		local sliderExtension = math.log(val/min)/(w*math.log(base))
		UiRect(w, 3)
		UiAlign("center middle")
		UiTranslate(-195, 1)
		val, done = UiSlider("ui/common/dot.png", "x", sliderExtension*w, 0, w) / w
		val = round((min*math.pow(base, val*195)), 2)
	UiPop()
	return val, done
end

function round(number, decimals)
    local power = 10^decimals
    return math.floor(number * power+0.5) / power
end