#include "scripts/default-values.lua"

function checkSetting(name, value, type)
	if not HasKey(name) then
		if type == "f" then
			SetFloat(name, value)
		elseif type == "b" then
			SetBool(name, value)
		elseif type == "s" then
			SetString(name, value)
		end
	end
end

function comma_value(amount)
  local formatted, k = amount, 0
  while true do  
    formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
    if (not k) or k==0 then
      break
    end
  end
  return formatted
end

onColor = {0, 1, 0}
offColor = {0.75, 0.75, 0.75}


function init()
	checkSetting("savegame.mod.showinfo", defaultValues["showInfo"], "b")
	checkSetting("savegame.mod.effectplayer", defaultValues["effectPlayer"], "b")
	checkSetting("savegame.mod.destroyclose", defaultValues["destroyClose"], "b")
	checkSetting("savegame.mod.growball", defaultValues["growBall"], "b")
	checkSetting("savegame.mod.maxtime", defaultValues["maxTime"], "f")
	checkSetting("savegame.mod.maxdist", defaultValues["maxDistance"], "f")
	checkSetting("savegame.mod.startmass", defaultValues["startMass"], "f")
	checkSetting("savegame.mod.maxmass", defaultValues["maxMass"], "f")
	checkSetting("savegame.mod.growspeed", defaultValues["growSpeed"], "f")
	checkSetting("savegame.mod.deletemass", defaultValues["deleteMass"], "f")
	checkSetting("savegame.mod.deletedist", defaultValues["deleteRadius"], "f")
	checkSetting("savegame.mod.menukey", defaultValues["menuKey"], "s")
	checkSetting("savegame.mod.killkey", defaultValues["killKey"], "s")
	checkSetting("savegame.mod.holename", defaultValues["holeName"], "s")
	
	bDestroyClose = GetBool("savegame.mod.destroyclose")
	bGrowBall = GetBool("savegame.mod.growball")
	bShowInfo = GetBool("savegame.mod.showinfo")
	bEffectPlayer = GetBool("savegame.mod.effectplayer")
	bMaxTime = GetFloat("savegame.mod.maxtime")
	bMaxDist = GetFloat("savegame.mod.maxdist")
	bStartMass = GetFloat("savegame.mod.startmass")
	bMaxMass = GetFloat("savegame.mod.maxmass")
	bGrowSpeed = GetFloat("savegame.mod.growspeed")
	bDeleteMass = GetFloat("savegame.mod.deletemass")
	bDeleteDist = GetFloat("savegame.mod.deletedist")
	bMenuKey = GetString("savegame.mod.menukey")
	bKillKey = GetString("savegame.mod.killkey")
	toggleSound = LoadSound("MOD/snd/toggle.ogg")
end

function draw()
	UiTranslate(UiCenter(), 100)
	UiAlign("center middle")

	UiFont("bold.ttf", 48)
	UiText("Omega Holes")
	UiTranslate(0, 50)
	UiFont("regular.ttf", 26)
	UiColor(0, 1, 0)
	UiText('You can edit these Parameters on the fly by right click or pressing "o" (Default)')
	UiTranslate(0, 30)
	UiColor(0.9, 0.9, 0.9)
	UiText("All Holes use these Parameters together")
	UiTranslate(0, 30)
	UiColor(1, 1, 0)
	UiText('Remove all Holes by pressing "p" (Default)')
	UiColor(0.9, 0.9, 0.9)
	UiTranslate(0, 50)
	UiPush()
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if UiTextButton("Reset to Default", 200, 40) then
			SetBool("savegame.mod.showinfo", defaultValues["showInfo"])
			SetBool("savegame.mod.effectplayer", defaultValues["effectPlayer"])
			SetBool("savegame.mod.destroyclose", defaultValues["destroyClose"])
			SetBool("savegame.mod.growball", defaultValues["growBall"])
			SetFloat("savegame.mod.maxtime", defaultValues["maxTime"])
			SetFloat("savegame.mod.maxdist", defaultValues["maxDistance"])
			SetFloat("savegame.mod.startmass", defaultValues["startMass"])
			SetFloat("savegame.mod.maxmass", defaultValues["maxMass"])
			SetFloat("savegame.mod.growspeed", defaultValues["growSpeed"])
			SetFloat("savegame.mod.deletemass", defaultValues["deleteMass"])
			SetFloat("savegame.mod.deletedist", defaultValues["deleteRadius"])
			SetString("savegame.mod.menukey", defaultValues["menuKey"])
			SetString("savegame.mod.killkey", defaultValues["killKey"])
			bShowInfo = defaultValues["showInfo"]
			bEffectPlayer = defaultValues["effectPlayer"]
			bDestroyClose = defaultValues["destroyClose"]
			bGrowBall = defaultValues["growBall"]
			bMaxTime = defaultValues["maxTime"]
			bMaxDist = defaultValues["maxDistance"]
			bStartMass = defaultValues["startMass"]
			bMaxMass = defaultValues["maxMass"]
			bGrowSpeed = defaultValues["growSpeed"]
			bDeleteMass = defaultValues["deleteMass"]
			bDeleteDist = defaultValues["deleteRadius"]
			bMenuKey = defaultValues["menuKey"]
			bKillKey = defaultValues["killKey"]

			InputEnabled_menukey = false
			InputEnabled_killkey = false
			PlaySound(toggleSound)
		end
	UiPop()

	UiTranslate(-350, 80)
	UiPush()
		UiAlign("left")
		UiText("Show Info Above")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.4, 0.9, 1)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if bShowInfo then
			UiColor(onColor[1], onColor[2], onColor[3])
			if UiTextButton("Yes", 60, 30) then
				bShowInfo = false
				SetBool("savegame.mod.showinfo", bShowInfo)
				PlaySound(toggleSound)
			end
		else
			UiColor(offColor[1], offColor[2], offColor[3])
			if UiTextButton("No", 60, 30) then
				bShowInfo = true
				SetBool("savegame.mod.showinfo", bShowInfo)
				PlaySound(toggleSound)
			end
		end
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Effect Player")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.4, 0.9, 1)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if bEffectPlayer then
			UiColor(onColor[1], onColor[2], onColor[3])
			if UiTextButton("Yes", 60, 30) then
				bEffectPlayer = false
				SetBool("savegame.mod.effectplayer", bEffectPlayer)
				PlaySound(toggleSound)
			end
		else
			UiColor(offColor[1], offColor[2], offColor[3])
			if UiTextButton("No", 60, 30) then
				bEffectPlayer = true
				SetBool("savegame.mod.effectplayer", bEffectPlayer)
				PlaySound(toggleSound)
			end
		end
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Destroy at Center")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.4, 0.9, 1)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if bDestroyClose then
			UiColor(onColor[1], onColor[2], onColor[3])
			if UiTextButton("Yes", 60, 30) then
				bDestroyClose = false
				SetBool("savegame.mod.destroyclose", bDestroyClose)
				PlaySound(toggleSound)
			end
		else
			UiColor(offColor[1], offColor[2], offColor[3])
			if UiTextButton("No", 60, 30) then
				bDestroyClose = true
				SetBool("savegame.mod.destroyclose", bDestroyClose)
				PlaySound(toggleSound)
			end
		end
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Grow the Ball")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.4, 0.9, 1)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if bGrowBall then
			UiColor(onColor[1], onColor[2], onColor[3])
			if UiTextButton("Yes", 60, 30) then
				bGrowBall = false
				SetBool("savegame.mod.growball", bGrowBall)
				PlaySound(toggleSound)
			end
		else
			UiColor(offColor[1], offColor[2], offColor[3])
			if UiTextButton("No", 60, 30) then
				bGrowBall = true
				SetBool("savegame.mod.growball", bGrowBall)
				PlaySound(toggleSound)
			end
		end
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Hole Life Time")
		UiTranslate(450, 0)
		UiAlign("right")
		bMaxTime, done = exponentialSlider(bMaxTime, 0.1, 7200)
		bMaxTime = round(bMaxTime, 2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		UiText(bMaxTime.." s")
		SetFloat("savegame.mod.maxtime", bMaxTime)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Max Pull Distance")
		UiTranslate(450, 0)
		UiAlign("right")
		bMaxDist, done = exponentialSlider(bMaxDist, 5, 500)
		bMaxDist = round(bMaxDist, 0)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		if bMaxDist > 499 then
			UiText("infinity m")
		else
			UiText(bMaxDist.." m")
		end
		SetFloat("savegame.mod.maxdist", bMaxDist)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Start Mass")
		UiTranslate(450, 0)
		UiAlign("right")
		bStartMass, done = exponentialSlider(bStartMass, 1000, 21000000000)
		bStartMass = round(bStartMass, -2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		UiText(comma_value(bStartMass/1000).." t")
		SetFloat("savegame.mod.startmass", bStartMass)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Max Mass")
		UiTranslate(450, 0)
		UiAlign("right")
		bMaxMass, done = exponentialSlider(bMaxMass, 1000, 21000000000)
		bMaxMass = round(bMaxMass, -2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		UiText(comma_value(bMaxMass/1000).." t")
		SetFloat("savegame.mod.maxmass", bMaxMass)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Mass Grow in % per s")
		UiTranslate(450, 0)
		UiAlign("right")
		bGrowSpeed, done = exponentialSlider(bGrowSpeed, 0.25, 1000)
		bGrowSpeed = round(bGrowSpeed, 2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		UiText(bGrowSpeed.." %/s")
		SetFloat("savegame.mod.growspeed", bGrowSpeed)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Remove Debris below")
		UiTranslate(450, 0)
		UiAlign("right")
		bDeleteMass, done = exponentialSlider(bDeleteMass, 0.01, 1000)
		bDeleteMass = round(bDeleteMass, 2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 9.0)
		UiText(bDeleteMass.." kg")
		SetFloat("savegame.mod.deletemass", bDeleteMass)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Hole Radius")
		UiAlign("right")
		UiTranslate(450, 0)
		UiAlign("right")
		bDeleteDist, done = exponentialSlider(bDeleteDist, 0.1, 100)
		bDeleteDist = round(bDeleteDist, 2)
		UiTranslate(210, 0)
		UiColor(0.9, 0.9, 0.9)
		UiText(bDeleteDist.." m")
		SetFloat("savegame.mod.deletedist", bDeleteDist)
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Menu Key")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.9, 0.9, 0.9)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if InputEnabled_menukey then
			UiColor(0, 1, 0)
			local currentM_key = InputLastPressedKey()
			if currentM_key ~= "" and currentM_key ~= "esc" then
				InputEnabled_menukey = false
				bMenuKey = currentM_key
				SetString("savegame.mod.menukey", currentM_key)
				PlaySound(toggleSound)
			end
		end
		local currentM_value = GetString("savegame.mod.menukey")
		if UiTextButton(currentM_value, 60, 30) then
			InputEnabled_menukey = true
			bMenuKey = "O"
			SetString("savegame.mod.menukey", "O")
		end
	UiPop()

	UiTranslate(0, 50)
	UiPush()
		UiAlign("left")
		UiText("Hole Kill Key")
		UiTranslate(310, 0)
		UiAlign("right")
		UiColor(0.9, 0.9, 0.9)
		UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)
		if InputEnabled_killkey then
			UiColor(0, 1, 0)
			local currentK_key = InputLastPressedKey()
			if currentK_key ~= "" and currentK_key ~= "esc" then
				InputEnabled_killkey = false
				bKillKey = currentK_key
				SetString("savegame.mod.killkey", currentK_key)
				PlaySound(toggleSound)
			end
		end
		local currentK_value = GetString("savegame.mod.killkey")
		if UiTextButton(currentK_value, 60, 30) then
			InputEnabled_killkey = true
			bKillKey = "V"
			SetString("savegame.mod.killkey", "V")
		end
	UiPop()

	UiButtonImageBox("ui/common/box-outline-6.png", 6, 6)

	UiTranslate(350, 50)
	if UiTextButton("Close", 120, 40) then
		PlaySound(toggleSound)
		Menu()
	end
end


function exponentialSlider(val, min, max)
	UiColor(.9, 0.7, 0)
	UiPush()
		UiTranslate(0, -8)
		local w = 195
		local done = false
		local base = math.pow(max/min, 1/w)
		local sliderExtension = math.log(val/min)/(w*math.log(base))
		UiRect(w, 4)
		UiAlign("center middle")
		UiTranslate(-w, 1)
		val, done = UiSlider("ui/common/dot.png", "x", sliderExtension*w, 0, w) / w
		val = round((min*math.pow(base, val*w)), 2)
	UiPop()
	return val, done
end

function round(number, decimals)
    local power = 10^decimals
    return math.floor(number * power+0.5) / power
end