function normalGravity(direction, distance, currentHole, b)
	return direction, currentHole.gravityAmplitude
end

function downGravity(direction, distance, currentHole, b)
	return Vec(0, -distance, 0), currentHole.gravityAmplitude
end

function sideGravity(direction, distance, currentHole, b)
	return VecScale(currentHole.shotDirection, distance), currentHole.gravityAmplitude
end

function spinGravity(direction, distance, currentHole, b)
	return QuatRotateVec(QuatEuler(0, 35, 0), direction), currentHole.gravityAmplitude
end

function bubbleGravity(direction, distance, currentHole, b)
	return direction, currentHole.gravityAmplitude * (distance - currentHole.sizeFactor * currentHole.initialDeleteRadius * 3)/distance
end

function vortexGravity(direction, distance, currentHole, b)
	if distance%20 < 10 then
		return QuatRotateVec(QuatEuler(0, 70, 0), direction), currentHole.gravityAmplitude
	else
		return QuatRotateVec(QuatEuler(0, -250, 0), direction), -currentHole.gravityAmplitude
	end
end

function pulsGravity(direction, distance, currentHole, b)
	return Vec(0, -distance, 0), currentHole.gravityAmplitude * currentHole.sinFactor+0.25
end

function waveGravity(direction, distance, currentHole, b)
	return direction, currentHole.gravityAmplitude * currentHole.sinFactor+0.125
end

function timeGravity(direction, distance, currentHole, b)
	if b == nil then
		velVec = GetPlayerVelocity()
	else
		velVec = GetBodyVelocity(b)
	end
	if VecLength(velVec) < 0.1 then
		return Vec(0,0,0), currentHole.gravityAmplitude
	end
	return VecScale(VecNormalize(velVec), distance), currentHole.gravityAmplitude*0.1
end

function spiralGravity(direction, distance, currentHole, b)
	return QuatRotateVec(QuatEuler(0, math.min(distance, 89), 0), direction), currentHole.gravityAmplitude
end

function twistGravity(direction, distance, currentHole, b)
	zangle = -90
	if direction[3] < 0 then
		zangle = -zangle
	end
	xangle = 90
	if direction[1] < 0 then
		xangle = -xangle
	end
	return QuatRotateVec(QuatEuler(0, 0, xangle), direction), currentHole.gravityAmplitude
end

local holeTemplateConfig = {
	{
		name = "blackhole",
		gravityDirection = normalGravity,
		gravityAmplitude = 1,
		r = 0,
		g = 0,
		b = 0,
	},
	{
		name = "greyhole",
		gravityDirection = waveGravity,
		gravityAmplitude = 1,
		r = 0.39,
		g = 0.39,
		b = 0.39,
	},
	{
		name = "whitehole",
		gravityDirection = normalGravity,
		gravityAmplitude = -1,
		r = 1,
		g = 1,
		b = 1,
	},
	{
		name = "uphole",
		gravityDirection = downGravity,
		gravityAmplitude = -1,
		r = 0,
		g = 1,
		b = 1,
	},
	{
		name = "wavehole",
		gravityDirection = pulsGravity,
		gravityAmplitude = 1,
		r = 0.5,
		g = 0,
		b = 1,
	},
	{
		name = "downhole",
		gravityDirection = downGravity,
		gravityAmplitude = 1,
		r = 1,
		g = 0,
		b = 0,
	},
	{
		name = "sidehole",
		gravityDirection = sideGravity,
		gravityAmplitude = 1,
		r = 0,
		g = 0,
		b = 1,
	},
	{
		name = "spinhole",
		gravityDirection = spinGravity,
		gravityAmplitude = 1,
		r = 1,
		g = 1,
		b = 0,
	},
	{
		name = "bubblehole",
		gravityDirection = bubbleGravity,
		gravityAmplitude = 1,
		r = 0,
		g = 1,
		b = 0,
	},
	{
		name = "vortexhole",
		gravityDirection = vortexGravity,
		gravityAmplitude = 1,
		r = 1,
		g = 0,
		b = 1,
	},
	{
		name = "timehole",
		gravityDirection = timeGravity,
		gravityAmplitude = -1,
		r = 1,
		g = 0.5,
		b = 0,
	},
	{
		name = "speedhole",
		gravityDirection = timeGravity,
		gravityAmplitude = 1,
		r = 0.5,
		g = 1,
		b = 0,
	},
	-- {
	-- 	name = "spiralhole",
	-- 	gravityDirection = spiralGravity,
	-- 	gravityAmplitude = 1,
	-- 	r = 0,
	-- 	g = 1,
	-- 	b = 0.5,
	-- },
	-- {
	-- 	name = "twisthole",
	-- 	gravityDirection = twistGravity,
	-- 	gravityAmplitude = 1,
	-- 	r = 0,
	-- 	g = 0.5,
	-- 	b = 1,
	-- }

}
_G.holeTemplateConfig = holeTemplateConfig

