#include "hole-config.lua"
#include "default-values.lua"
local holeTemplateConfig = holeTemplateConfig
_G.holeTemplateConfig = nil

holeHandler = {
	holeNum = 1,
	holes = {}
}



function loadConfig()
	local holes = {}
	for i=1, #holeTemplateConfig do
		local firsthole = holeTemplateConfig[i]
		firsthole.sound = LoadLoop("MOD/snd/"..firsthole.name..".ogg", 15)
		holes[firsthole.name] = firsthole
	end
	return holes
end

function tobool(str)
    if str == nil then
        return false
    end
    return string.lower(str) == "true"
end


function GetValue(body, tag, type)
	local val = GetTagValue(body, tag)
	if val == "" then
		return defaultValues[tag]
	else
		if type == "s" then
			return val
		elseif type == "f" then
			return tonumber(val)
		elseif type == "b" then
			return tobool(val)
		end
	end
end

function handleCommand(cmd)
	if cmd == "quickload" then
		holeTemplates = loadConfig()
		for key, currentHole in ipairs(holeHandler.holes) do
			currentHole.gravityDirection = holeTemplates[currentHole.name].gravityDirection
		end
	end
end


function createHole()
	local holeBody = FindBody("omegahole")
	holeName = GetValue(holeBody, "holeName", "s")
	showInfo = GetValue(holeBody, "showInfo", "b")
	effectPlayer = GetValue(holeBody, "effectPlayer", "b")
	destroyClose = GetValue(holeBody, "destroyClose", "b")
	growBall = GetValue(holeBody, "growBall", "b")
	maxTime = GetValue(holeBody, "maxTime", "f")
	maxDistance = GetValue(holeBody, "maxDistance", "f")
	startMass = GetValue(holeBody, "startMass", "f")
	maxMass = GetValue(holeBody, "maxMass", "f")
	growSpeed = GetValue(holeBody, "growSpeed", "f")
	deleteMass = GetValue(holeBody, "deleteMass", "f")
	deleteRadius = GetValue(holeBody, "deleteRadius", "f")
	frequency = GetValue(holeBody, "frequency", "f")
	deleteRadius = math.max(deleteRadius, 1)

	holeHandler.holes[holeHandler.holeNum] = deepcopy(holeTemplates[holeName])
	local currentHole = holeHandler.holes[holeHandler.holeNum]
	currentHole.name = holeName
	currentHole.showInfo = showInfo
	currentHole.effectPlayer = effectPlayer
	currentHole.destroyClose = destroyClose
	currentHole.growBall = growBall
	currentHole.maxTime = maxTime
	if maxDistance > 499 then
		currentHole.maxDistance = 1000000
	else
		currentHole.maxDistance = maxDistance
	end
	currentHole.maxMass = maxMass
	currentHole.growSpeed = growSpeed
	currentHole.deleteMass = deleteMass
	currentHole.deleteRadius = deleteRadius
	currentHole.initialDeleteRadius = deleteRadius
	currentHole.frequency = frequency

	currentHole.body = holeBody
	currentHole.pos = TransformToParentPoint(GetBodyTransform(currentHole.body), GetBodyCenterOfMass(currentHole.body))
	currentHole.inair = false
	currentHole.timeToLive = maxTime
	currentHole.mass = startMass
	currentHole.dist = 0
	if currentHole.destroyClose then
		MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
	end
	if currentHole.growBall then
		spawnBall(currentHole)
	end
end


function init()
	holeTemplates = loadConfig()
	createHole(holeName)
	gamma = 6.6743*10^-4
end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end


function spawnBall(currentHole)
	local ballPos = VecCopy(currentHole.pos)
	ballPos[2] = ballPos[2] - currentHole.deleteRadius/2
	currentHole.ball = Spawn(string.format('<vox tag="unbreakable dynamic=false physical=false" physical="false" dynamic="false" collide="false" file="MOD/vox/holes/%s.vox" scale="%.2f"/>', currentHole.name, currentHole.deleteRadius/5), Transform(ballPos), true)[1]
end

function replaceBall(currentHole)
	local b = currentHole.ball
	if IsHandleValid(b) then
		Delete(b)
	end
	spawnBall(currentHole)
end

function Hole(dt, currentHole)
	if currentHole.effectPlayer then
		GravityEffectPlayer(dt, currentHole)
	end
	if currentHole.destroyClose then
		if math.random() < 0.3 then
			RipAndTear(currentHole)
		end
	end
	if currentHole.growBall then
		tickBall(currentHole)
	end
	local mi = VecAdd(currentHole.pos, Vec(-currentHole.dist, -currentHole.dist, -currentHole.dist))
	local ma = VecAdd(currentHole.pos, Vec(currentHole.dist, currentHole.dist, currentHole.dist))
	QueryRequire("physical dynamic")
	local bodies = QueryAabbBodies(mi, ma)
	for i=1,#bodies do
		local b = bodies[i]
		if GetTagValue(b, "omegaattract") ~= "no" then
			local mass = GetBodyMass(b)
			if mass < currentHole.deleteMass then
				if GetTagValue(b, "omegadestroy") ~= "no" then
					Delete(b)
					currentHole.mass = math.min(currentHole.maxMass, currentHole.mass + mass)
				end
			else
				local bc = TransformToParentPoint(GetBodyTransform(b), GetBodyCenterOfMass(b))
				local dir = VecSub(currentHole.pos, bc)
				local dist = VecLength(dir)
					
				if dist < currentHole.deleteRadius then
					if GetTagValue(b, "omegadestroy") ~= "no" then
						Delete(b)
						currentHole.mass = math.min(currentHole.maxMass, currentHole.mass + mass)
					end
				elseif dist < currentHole.dist then
					local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole, b)
					local impulse = VecScale(gdir, dt * amplitude * gamma * mass * currentHole.mass/(dist * dist * dist))
					ApplyBodyImpulse(b, bc, impulse)
				end
			end
		end
	end
end



function tickBall(currentHole)
	local ballRadius = currentHole.initialDeleteRadius * currentHole.sizeFactor
	if ballRadius > currentHole.deleteRadius * 1.15 and ballRadius < 100 then
		currentHole.deleteRadius = ballRadius
		MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
		replaceBall(currentHole)
	end
end


function GetRandomVector(OffsetRadius, currentHoleWidth)
	local lenVec = Vec(0, math.random()*currentHoleWidth+OffsetRadius, 0)
	return QuatRotateVec(QuatEuler(math.random(360), math.random(360), math.random(360)), lenVec)
end

function RipAndTear(currentHole)
	local rippingRadius = math.min(currentHole.maxDistance-1, math.sqrt(currentHole.mass * gamma/100))
	if rippingRadius > currentHole.deleteRadius then
		local holePosition = VecAdd(currentHole.pos, GetRandomVector(currentHole.deleteRadius, rippingRadius-currentHole.deleteRadius))
		local scaledGForce = math.min(math.abs(GetGForce(currentHole, holePosition)), 400)
		MakeHole(holePosition, scaledGForce, math.max(0, 0.75*scaledGForce-15), math.max(0, 0.5*scaledGForce-20))
	end
end

function GravityEffectPlayer(dt, currentHole)
	local pt = GetPlayerPos()
	local dir = VecSub(currentHole.pos, pt)
	local dist = VecLength(dir)
	if dist < currentHole.dist then
		if dist < 0.4*currentHole.deleteRadius then
			RespawnPlayer()
		else
			if dist < currentHole.deleteRadius then
				SetPlayerHealth(math.min(GetPlayerHealth()-0.1*dist/currentHole.deleteRadius, 0))
			end
			local pv = GetPlayerVelocity()
			local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
			local add = VecScale(gdir, dt * amplitude * gamma * currentHole.mass/(dist * dist * dist))
			SetPlayerVelocity(VecAdd(pv, add))
		end
	end
end




function endHole(currentHole)
	MakeHole(currentHole.pos, currentHole.deleteRadius*2, currentHole.deleteRadius*1.5, currentHole.deleteRadius)
	local b = currentHole.ball
	if IsHandleValid(b) then
		Delete(b)
	end
	if currentHole.showInfo then
		if currentHole.timeToLive < 0.01 then
			SetString("hud.notification", (currentHole.name:gsub("^%l", string.upper)).." Vaporized")
		else
			SetString("hud.notification", (currentHole.name:gsub("^%l", string.upper)).." Removed")
		end
	end
	currentHole.timeToLive = 0
end

function GetGForce(currentHole, position)
	local dir = VecSub(currentHole.pos, position)
	local dist = VecLength(dir)
	
	if dist < currentHole.dist then
		local gdir, amplitude = currentHole.gravityDirection(dir, dist, currentHole)
		return amplitude * gamma * currentHole.mass/(dist * dist * 10)
	else
		return 0
	end
end

function round(number, decimals)
    local power = 10^decimals
    return math.floor(number * power+0.5) / power
end

function comma_value(amount)
  local formatted, k = amount, 0
  while true do  
    formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
    if (not k) or k==0 then
      break
    end
  end
  return formatted
end

function tick(dt)
	for key, currentHole in ipairs(holeHandler.holes) do
		currentHole.pos = TransformToParentPoint(GetBodyTransform(currentHole.body), GetBodyCenterOfMass(currentHole.body))
		if currentHole.timeToLive > 0 then
			currentHole.timeToLive = currentHole.timeToLive - dt
			if currentHole.timeToLive < 0.01 or (not IsHandleValid(currentHole.body)) then
				endHole(currentHole)
				holeHandler.holes[key] = {timeToLive=0}
			else
				if currentHole.mass < currentHole.maxMass then
					currentHole.mass = math.min(currentHole.mass * (1+dt*0.00727*currentHole.growSpeed), currentHole.maxMass)
				end
				if currentHole.dist < currentHole.maxDistance then
					currentHole.dist = math.min(currentHole.maxDistance, math.sqrt(currentHole.mass * gamma/0.5))
				end
				currentHole.shotDirection = QuatRotateVec(GetBodyTransform(currentHole.body).rot, Vec(0, 0, 1))
				currentHole.sizeFactor = (20*math.pow(currentHole.mass, 1/2)/math.pow(2100000000, 1/2))+1
				currentHole.sinFactor = math.sin(currentHole.timeToLive*currentHole.frequency*math.pi)
				PlayLoop(currentHole.sound, currentHole.pos, 0.5*math.log(currentHole.mass)/math.log(currentHole.maxMass))
				if currentHole.showInfo then
					SetString("hud.notification", string.format("Mass "..comma_value(round(currentHole.mass/1000, 0)).." t - Gravity "..comma_value(round(GetGForce(currentHole, GetPlayerPos()), 2)).." g"))
				end
				Hole(dt, currentHole)
			end
		end
	end
end